
// a simple driver for the simfold

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

// include the simfold header files
#include "simfold.h"
#include "externs.h"
#include "constants.h"

int main (int argc, char *argv[])
{
    char sequence[MAXSLEN];
    char structure[MAXSLEN];
    char restricted[MAXSLEN];
    double energy;
    char structures[MAXSUBSTR][MAXSLEN];
    double energies[MAXSUBSTR];     
    
    if (argc != 2 && argc != 3)
    {
        printf ("Usage: %s <sequence> [<restricted_structure>]\n", argv[0]);
        printf ("Example: %s \"GAAACGAUGACAUAC\"\n", argv[0]);
        printf ("Example: %s \"GAAACGAUGACAUAC\" \"(_____________)\"\n", argv[0]);
        return 0;
    }        
    strcpy (sequence, argv[1]);    
    if (argc == 3)
        strcpy (restricted, argv[2]);
    
    // Before calling any function in the library, you have to initialize config_file, dna_or_rna, temperature
    //     and to call the function init_data, which loads the thermodynamic parameters into memory
    
    // configuration file, the path should relative to the location of this executable
    char config_file[200] = "params/multirnafold.conf";

    // what to fold: RNA or DNA
    int dna_or_rna = RNA;

    // temperature: any integer or real number between 0 and 100
    // represents degrees Celsius
    double temperature = 37; 
    
    // initialize the thermodynamic parameters
    // call init_data only once for the same dna_or_rna and same temperature
    // if one of them changes, call init_data again   
    init_data (argv[0], config_file, dna_or_rna, temperature);         

    energy = simfold (sequence, structure);
    printf ("Seq: %s\nMFE: %s  %.2lf\n", sequence, structure, energy);
    
    // if this structure must be restricted
    if (argc == 3)
    {
        // you can return only the MFE-restricted structure, or several (all) structures having MFE-restricted
        //energy = simfold_restricted (sequence, restricted, structure);
        //printf ("RES: %s  %.2lf\n", structure, energy);
        
        // Return all structures having the mfe
        int num_str = simfold_restricted_all_mfe_structures (sequence, restricted, structures, energies);
        for (int i=0; i < num_str; i++)
        {
            printf ("RES: %s  %.2lf\n", structures[i], energies[i]);
        }
    }    
    
    return 0;
}
    


