/***************************************************************************
                          params.h  -  description
                             -------------------
    begin                : Sun May 8 2005
    copyright            : (C) 2005 by Mirela Andronescu
    email                : andrones@cs.ubc.ca
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef PARAMS_H
#define PARAMS_H


void print_stacking_energies();
// prints the stacking energies

void print_tstacki_dangling_energies();
// prints the tstacki and dangling energies

void print_stack_dangling_energies();
// prints the stack and dangling energies

void print_stack_equation_dangling();
// prints the stack and dangling energies

void print_int22_tstacki();
// prints the int22 and tstacki energies


void count_AU_penalty (int base_i, int base_j, int *counter);
// PRE:  base_i and base_j make a pair
// POST: increment counter at the appropriate position
// Mirela: Nov 23, 2003

void count_penalty_by_size (int size, char type, int *counter);
// PRE:  size is the size of the loop
//       type is HAIRP or INTER or BULGE
// POST: return the penalty by size of the loop
// Mirela: Nov 23, 2003

int structure_type_index (char type[]);
  // Mirela: Nov 23, 2003
  // Given the type as a string, return the index in string_params

void count_asymmetry_penalty (int size1, int size2, int *counter);
// PRE:  size1 and size2 are the sizes of the two free base groups
// POST: Calculate the asymmetry penalty for internal loops
//       Note that if size1 == size2, pen is 0
// Mirela: Nov 23, 2003

void save_parameters (char *filename);
  // Mirela: Dec 16, 2003
  // save all parameters in the given file

int create_building_block_strings ();
// Mirela: Sep 20, 2005
// For most of the building blocks, write which sequence(s) correspond to each building block
// also fill string_params
  
int create_string_params ();
  // Mirela: Nov 23, 2003
  // writes each parameter type, excluding duplicates, in a long vector, containing the names of the parameters

void count_each_structure_type (char *sequence, char *structure, int *counter);
// Mirela: Nov 23, 2003
// GIven the sequence and the structure, returns the vector counter, with the #of elementary structures


double simfold_restricted_logZ (char *sequence, char *real_structure, char *restricted, double &min_energy, double &max_energy, int &actual_num_str);

void fill_data_structures_with_new_parameters (char *filename);
  // Mirela: Dec 16, 2003
  // reads parameters from a file, and writes them in the internal data structures
  // PRE: first read the actual standard parameters, to be able to figure out which of them are
  // < INF, and maybe to also keep some old values.

double compute_f (char *input_file);
// PRE:  the parameters have been read
// POST: computes the f function, which is sum_{i=1}^N (1/RT G_{i,nat,theta} + logZ)

void compute_gradient_f (char *input_file, double *f_gradient);
// PRE:  the parameters have been read
// POST: compute the gradient of f, which is 1/RT sum_{i=1}^N (c_{nat}^i - sum_k c_k^i exp (-1/RT G_{i,k,theta}) / sum_k exp (-1/RT G_{i,k,theta}))

double compute_f_and_gradient_f (char *input_file, double *f_gradient);
// PRE:  the parameters have been read
// POST: computes the f function, which is sum_{i=1}^N (1/RT G_{i,nat,theta} + logZ)
// POST: compute the gradient of f, which is 1/RT sum_{i=1}^N (c_{nat}^i - sum_k c_k^i exp (-1/RT G_{i,k,theta}) / sum_k exp (-1/RT G_{i,k,theta}))

void compute_counts_vector_LP (char *input_file, int *total_counter);
// PRE:  the parameters have been read
// POST: given p input sequence/structure pairs, compute sum_{j=1}^p (c_j, cbar_j)
//          where c_j is the counts for the known structures
//          and cbar_j is the counts for the predicted mfe structure, with the given set of parameters

void compute_counts_matrix_LP (char *input_file, int train_samples);
// PRE:  the parameters have been read
//       train_samples = # of training instances. -1 if all.
// POST: given p input sequence/structure pairs, compute and display the counts c_j - cbar_j
//          where c_j is the counts for the known structures
//          and cbar_j is the counts for the predicted mfe structure, with the given set of parameters


#endif

